@@ -18,6 +18,7 @@
 #include "file_utils.h"
 #include "log.h"
 #include "lsm.h"
+#include "open_utils.h"
 #include "parse.h"
 #include "process_utils.h"
 #include "utils.h"
@@ -974,7 +975,7 @@ static int load_apparmor_profile(struct lsm_ops *ops, struct lxc_conf *conf, con
 		old_len = profile_sb.st_size;
 		old_content = lxc_strmmap(NULL, old_len, PROT_READ,
 		                          MAP_PRIVATE, profile_fd, 0);
-		if (!old_content) {
+		if (old_content == MAP_FAILED) {
 			SYSERROR("Failed to mmap old profile from %s",
 			         profile_path);
 			goto out;
