@@ -19,8 +19,6 @@
 #include <string.h>
 #include <sys/mman.h>
 #include <sys/mount.h>
-/* Needs to be after sys/mount.h header */
-#include <linux/fs.h>
 #include <sys/param.h>
 #include <sys/prctl.h>
 #include <sys/stat.h>
@@ -34,6 +32,7 @@
 #include "lxclock.h"
 #include "memory_utils.h"
 #include "namespace.h"
+#include "open_utils.h"
 #include "parse.h"
 #include "process_utils.h"
 #include "syscall_wrappers.h"
@@ -218,6 +217,9 @@ int mkdir_p(const char *dir, mode_t mode)
 	const char *tmp = dir;
 	const char *orig = dir;
 
+	if (access(dir, F_OK) != -1)
+		return 0;
+
 	do {
 		__do_free char *makeme = NULL;
 		int ret;
@@ -347,7 +349,7 @@ bool wait_exited(pid_t pid)
 	return true;
 }
 
-#ifdef HAVE_OPENSSL
+#if HAVE_OPENSSL
 #include <openssl/evp.h>
 
 static int do_sha1_hash(const char *buf, int buflen, unsigned char *md_value,
@@ -1094,7 +1096,7 @@ int __safe_mount_beneath_at(int beneath_fd, const char *src, const char *dst, co
 			    unsigned int flags, const void *data)
 {
 	__do_close int source_fd = -EBADF, target_fd = -EBADF;
-	struct lxc_open_how how = {
+	struct open_how how = {
 		.flags		= PROTECT_OPATH_DIRECTORY,
 		.resolve	= PROTECT_LOOKUP_BENEATH_WITH_MAGICLINKS,
 	};
@@ -1789,7 +1791,7 @@ bool lxc_can_use_pidfd(int pidfd)
 	int ret;
 
 	if (pidfd < 0)
-		return log_error(false, "Kernel does not support pidfds");
+		return log_trace(false, "Kernel does not support pidfds");
 
 	/*
 	 * We don't care whether or not children were in a waitable state. We
